from unittest import mock
from json.decoder import JSONDecodeError

import pytest
import requests

from civis.base import Endpoint, get_base_url, CivisAPIError


def test_base_url_default():
    assert get_base_url() == "https://api.civisanalytics.com/"


def test_base_url_from_env():
    custom_url = "https://api1.civisanalytics.com"
    with mock.patch.dict("os.environ", {"CIVIS_API_ENDPOINT": custom_url}):
        assert get_base_url() == custom_url + "/"


@mock.patch("civis.base.get_base_url", return_value="https://base.api.url/")
def test_endpoint_base_url(mock_get_base_url):
    session = mock.MagicMock(spec=requests.Session)
    endpoint = Endpoint(session, "client")

    assert endpoint._base_url == "https://base.api.url/"


def test_store_last_response():
    mock_client = mock.Mock()
    endpoint = Endpoint({}, client=mock_client, return_type="raw")

    returned_resp = {"value": "response"}
    endpoint._make_request = mock.Mock(return_value=returned_resp)

    resp = endpoint._call_api("GET")
    assert resp == returned_resp
    assert mock_client.last_response is resp


def test_civis_api_error_empty_response():
    # Fake response object, try to trigger error
    # Make sure response.json() gets the JSON decode error
    response = requests.Response()
    response._content = b"foobar"
    with pytest.raises(JSONDecodeError):
        response.json()

    error = CivisAPIError(response)
    assert error.error_message == "No Response Content from Civis API"


@pytest.mark.parametrize(
    "source_params, expected_params",
    [
        ({}, None),
        (None, None),
        (
            {"foo": 123, "bar": "hello", "baz": {"a": 1, "b": 2}},
            {"foo": 123, "bar": "hello", "baz": {"a": 1, "b": 2}},
        ),
        ({"foo": [1, 2, 3]}, {"foo[]": [1, 2, 3]}),
        ({"foo": (1, 2, 3)}, {"foo[]": [1, 2, 3]}),
        ({"foo": {1, 2, 3}}, {"foo[]": [1, 2, 3]}),
    ],
)
def test_array_params(source_params, expected_params):
    assert Endpoint._handle_array_params(source_params) == expected_params



import logging
import os

import tenacity
from tenacity.wait import wait_base


log = logging.getLogger(__name__)

_RETRY_CODES = [429, 502, 503, 504]
_RETRY_VERBS = ["HEAD", "TRACE", "GET", "PUT", "OPTIONS", "DELETE"]
_POST_RETRY_CODES = [429, 503]


# Defining the default tenacity.Retrying as a user-friendly code string
# so that it can be shown in civis.APIClient's docstring.
DEFAULT_RETRYING_STR = """
tenacity.Retrying(
    wait=tenacity.wait_random_exponential(multiplier=2, max=60),
    stop=(tenacity.stop_after_delay(600) | tenacity.stop_after_attempt(10)),
    retry_error_callback=lambda retry_state: retry_state.outcome.result(),
)
"""

# Explicitly set the available globals and locals
# to mitigate risk of unwanted code execution
DEFAULT_RETRYING = eval(  # nosec
    DEFAULT_RETRYING_STR,
    {"tenacity": tenacity, "__builtins__": {}},  # globals
    {},  # locals
)


def get_api_key(api_key):
    """Pass-through if `api_key` is not None otherwise tries the CIVIS_API_KEY
    environment variable.
    """
    if api_key is not None:  # always prefer user given one
        return api_key
    api_key = os.environ.get("CIVIS_API_KEY", None)
    if api_key is None:
        raise EnvironmentError(
            "No Civis API key found. Please store in "
            "CIVIS_API_KEY environment variable"
        )
    return api_key


def retry_request(method, prepared_req, session, retrying=None):
    retry_conditions = None
    retrying = retrying if retrying else DEFAULT_RETRYING

    def _make_request(req, sess):
        """send the prepared session request"""
        response = sess.send(req)
        return response

    if method.upper() == "POST":
        retry_conditions = tenacity.retry_if_result(
            lambda res: res.status_code in _POST_RETRY_CODES
        )
    elif method.upper() in _RETRY_VERBS:
        retry_conditions = tenacity.retry_if_result(
            lambda res: res.status_code in _RETRY_CODES
        )

    if retry_conditions:
        retrying.retry = retry_conditions
        retrying.wait = wait_for_retry_after_header(fallback=retrying.wait)
        response = retrying(_make_request, prepared_req, session)
        return response

    response = _make_request(prepared_req, session)
    return response


class wait_for_retry_after_header(wait_base):
    """Wait strategy that first looks for Retry-After header. If not
    present it uses the fallback strategy as the wait param"""

    def __init__(self, fallback):
        self.fallback = fallback

    def __call__(self, retry_state):
        # retry_state is an instance of tenacity.RetryCallState.
        # The .outcome property contains the result/exception
        # that came from the underlying function.
        result_headers = retry_state.outcome._result.headers
        retry_after = result_headers.get("Retry-After") or result_headers.get(
            "retry-after"
        )

        try:
            log.info("Retrying after {} seconds".format(retry_after))
            return int(retry_after)
        except (TypeError, ValueError):
            pass
        return self.fallback(retry_state)



import json
import os
import tempfile
import time
from collections import defaultdict, OrderedDict
from unittest import mock

import pytest
from jsonref import JsonRef
from requests.exceptions import HTTPError

from civis.base import DEFAULT_USER_AGENT, Endpoint
from civis.resources import _resources, API_SPEC_PATH
from civis.resources._client_pyi import generate_client_pyi, CLIENT_PYI_PATH
from civis.tests import create_client_mock


with open(API_SPEC_PATH) as f:
    API_SPEC = json.load(f, object_pairs_hook=OrderedDict)


RESPONSE_DOC = """Returns
-------
:class:`civis.Response`
    - id : int
        The ID of the credential.
    - name : str
        The name identifying the credential
    - type : str
        The credential's type.
    - username : str
        The username for the credential.
    - description : str
        A long description of the credential.
    - owner : str
        The name of the user who this credential belongs to.
    - remote_host_id : int
        The ID of the remote host associated with this credential.
    - remote_host_name : str
        The name of the remote host associated with this credential.
    - created_at : str (time)
        The creation time for this credential.
    - updated_at : str (time)
        The last modification time for this credential."""  # noqa: E122


def test_create_method_iterator_kwarg():
    args = [
        {"name": "limit", "in": "query", "required": False, "doc": ""},
        {"name": "page_num", "in": "query", "required": False, "doc": ""},
        {"name": "order", "in": "query", "required": False, "doc": ""},
        {"name": "order_by", "in": "query", "required": False, "doc": ""},
    ]
    method = _resources.create_method(
        args,
        "get",
        "mock_name",
        "/objects",
        "deprecation",
        "param_doc",
        "resp_doc",
        "return_annotation",
    )
    mock_endpoint = mock.MagicMock()

    method(mock_endpoint, iterator=True)
    mock_endpoint._call_api.assert_called_once_with(
        "get", "/objects", {}, {}, "deprecation", iterator=True
    )


def test_create_method_no_iterator_kwarg():

    # Test that dynamically-created function errors when an
    # unexpected "iterator" parameter is passed in
    args = [{"name": "id", "in": "query", "required": True, "doc": ""}]
    method = _resources.create_method(
        args,
        "get",
        "mock_name",
        "/objects",
        "deprecation",
        "param_doc",
        "resp_doc",
        "return_annotation",
    )
    mock_endpoint = mock.MagicMock()

    with pytest.raises(TypeError) as excinfo:
        method(mock_endpoint, id=202, iterator=True)

    assert "keyword argument" in str(excinfo.value)

    # Dynamic functions handle optional argument through a different
    # code path; verify that this also rejects unexpected arguments.
    args2 = [{"name": "foo", "in": "query", "required": False, "doc": ""}]
    method2 = _resources.create_method(
        args2,
        "get",
        "mock_name",
        "/objects",
        "deprecation",
        "param_doc",
        "resp_doc",
        "return_annotation",
    )
    mock_endpoint2 = mock.MagicMock()
    with pytest.raises(TypeError) as excinfo:
        method2(mock_endpoint2, iterator=True)

    assert "keyword argument" in str(excinfo.value)


def test_exclude_resource():
    include = "tables/"
    exclude = "feature_flags/"
    assert _resources.exclude_resource(exclude, "1.0")
    assert not _resources.exclude_resource(include, "1.0")


def test_property_type():
    prop = {"type": "array"}
    prop2 = {"type": "object"}
    prop3 = {"type": "string", "format": "date"}
    assert _resources.property_type(prop) == "List"
    assert _resources.property_type(prop2) == "dict"
    assert _resources.property_type(prop3) == "str (date)"
    assert _resources.property_type(prop3, get_format=False) == "str"


def test_name_and_type_doc():
    prop = {"type": "string"}
    x = _resources.name_and_type_doc("A", prop, 0, True)
    y = _resources.name_and_type_doc("A", prop, 1, True)
    z = _resources.name_and_type_doc("A", prop, 0, False)
    assert x == "a : str, optional"
    assert y == "    - a : str, optional"
    assert z == "a : str"


def test_docs_from_property():
    prop = {"type": "array"}
    prop2 = {"type": "object", "properties": {"A": prop}}
    x = _resources.docs_from_property("A", prop, {}, 0, True)
    y = _resources.docs_from_property("B", prop2, {}, 0, False)
    assert sorted(x) == sorted(["a : List, optional"])
    assert sorted(y) == sorted(["b : dict", "    - a : List"])


def test_docs_from_properties():
    props = {"A": {"type": "string"}, "B": {"type": "integer"}}
    x = _resources.docs_from_properties(props, 0)
    y = _resources.docs_from_properties(props, 1)
    assert sorted(x) == sorted(["a : str", "b : int"])
    assert sorted(y) == sorted(["    - a : str", "    - b : int"])


def test_deprecated_notice():
    deprecation_warning = "This endpoint is no longer supported"
    notice = _resources.deprecated_notice(deprecation_warning)

    assert ".. warning::" in notice
    assert deprecation_warning in notice


def test_deprecated_notice_handles_none():
    assert _resources.deprecated_notice(None) == ""


def test_doc_from_responses():
    responses = OrderedDict(
        [
            (
                "200",
                OrderedDict(
                    [
                        ("description", "success"),
                        (
                            "schema",
                            OrderedDict(
                                [
                                    ("type", "array"),
                                    (
                                        "items",
                                        OrderedDict(
                                            [
                                                ("type", "object"),
                                                (
                                                    "properties",
                                                    OrderedDict(
                                                        [
                                                            (
                                                                "id",
                                                                OrderedDict(
                                                                    [
                                                                        (
                                                                            "description",  # noqa: E501
                                                                            "The ID of the credential.",  # noqa: E501
                                                                        ),
                                                                        (
                                                                            "type",
                                                                            "integer",
                                                                        ),
                                                                    ]
                                                                ),
                                                            ),
                                                            (
                                                                "name",
                                                                OrderedDict(
                                                                    [
                                                                        (
                                                                            "description",  # noqa: E501
                                                                            "The name identifying the credential",  # noqa: E501
                                                                        ),
                                                                        (
                                                                            "type",
                                                                            "string",
                                                                        ),
                                                                    ]
                                                                ),
                                                            ),
                                                            (
                                                                "type",
                                                                OrderedDict(
                                                                    [
                                                                        (
                                                                            "description",  # noqa: E501
                                                                            "The credential's type.",  # noqa: E501
                                                                        ),
                                                                        (
                                                                            "type",
                                                                            "string",
                                                                        ),
                                                                    ]
                                                                ),
                                                            ),
                                                            (
                                                                "username",
                                                                OrderedDict(
                                                                    [
                                                                        (
                                                                            "description",  # noqa: E501
                                                                            "The username for the credential.",  # noqa: E501
                                                                        ),
                                                                        (
                                                                            "type",
                                                                            "string",
                                                                        ),
                                                                    ]
                                                                ),
                                                            ),
                                                            (
                                                                "description",
                                                                OrderedDict(
                                                                    [
                                                                        (
                                                                            "description",  # noqa: E501
                                                                            "A long description of the credential.",  # noqa: E501
                                                                        ),
                                                                        (
                                                                            "type",
                                                                            "string",
                                                                        ),
                                                                    ]
                                                                ),
                                                            ),
                                                            (
                                                                "owner",
                                                                OrderedDict(
                                                                    [
                                                                        (
                                                                            "description",  # noqa: E501
                                                                            "The name of the user who this credential belongs to.",  # noqa: E501
                                                                        ),
                                                                        (
                                                                            "type",
                                                                            "string",
                                                                        ),
                                                                    ]
                                                                ),
                                                            ),
                                                            (
                                                                "remoteHostId",
                                                                OrderedDict(
                                                                    [
                                                                        (
                                                                            "description",  # noqa: E501
                                                                            "The ID of the remote host associated with this credential.",  # noqa: E501
                                                                        ),
                                                                        (
                                                                            "type",
                                                                            "integer",
                                                                        ),
                                                                    ]
                                                                ),
                                                            ),
                                                            (
                                                                "remoteHostName",
                                                                OrderedDict(
                                                                    [
                                                                        (
                                                                            "description",  # noqa: E501
                                                                            "The name of the remote host associated with this credential.",  # noqa: E501
                                                                        ),
                                                                        (
                                                                            "type",
                                                                            "string",
                                                                        ),
                                                                    ]
                                                                ),
                                                            ),
                                                            (
                                                                "createdAt",
                                                                OrderedDict(
                                                                    [
                                                                        (
                                                                            "description",  # noqa: E501
                                                                            "The creation time for this credential.",  # noqa: E501
                                                                        ),
                                                                        (
                                                                            "type",
                                                                            "string",
                                                                        ),
                                                                        (
                                                                            "format",
                                                                            "time",
                                                                        ),
                                                                    ]
                                                                ),
                                                            ),
                                                            (
                                                                "updatedAt",
                                                                OrderedDict(
                                                                    [
                                                                        (
                                                                            "description",  # noqa: E501
                                                                            "The last modification time for this credential.",  # noqa: E501
                                                                        ),
                                                                        (
                                                                            "type",
                                                                            "string",
                                                                        ),
                                                                        (
                                                                            "format",
                                                                            "time",
                                                                        ),
                                                                    ]
                                                                ),
                                                            ),
                                                        ]
                                                    ),
                                                ),
                                            ]
                                        ),
                                    ),
                                ]
                            ),
                        ),
                    ]
                ),
            )
        ]
    )  # noqa: E501
    x = _resources.doc_from_responses(responses, False, False)
    assert x == RESPONSE_DOC


def test_type_from_param():
    params = [
        {"type": "string"},
        {"type": "integer"},
        {"type": "array"},
        {"type": "array", "items": {"type": "integer"}},
        {"type": "array", "items": {"$ref": "#/definitions/Object0"}},
        {"type": "object"},
    ]
    assert _resources.type_from_param(params[0]) == "str"
    assert _resources.type_from_param(params[1]) == "int"
    assert _resources.type_from_param(params[2]) == "List"
    assert _resources.type_from_param(params[3]) == "List[int]"
    assert _resources.type_from_param(params[4]) == "List[dict]"
    assert _resources.type_from_param(params[4], skip_dict_item_type=True) == "List"
    assert _resources.type_from_param(params[5]) == "dict"
    assert (
        _resources.type_from_param(params[5], in_returned_object=True)
        == ":class:`civis.Response`"
    )  # noqa: E501


def test_iterable_method():
    assert _resources.iterable_method("get", ["limit", "page_num"])
    assert not _resources.iterable_method("get", ["page_num"])
    assert not _resources.iterable_method("post", ["limit", "page_num"])


def test_split_method_params():
    params = [
        {"name": "a", "required": True, "in": "body", "type": "integer"},
        {"name": "b", "required": True, "in": "path", "type": "integer"},
        {"name": "c", "required": True, "in": "query", "type": "string"},
        {"name": "d", "required": False, "in": "query", "type": "string"},
    ]
    x = _resources.split_method_params(params)
    args, kwargs, body_params, query_params, path_params = x
    assert sorted(args.keys()) == sorted(["a", "b", "c"])
    assert args["a"] == {"type": "integer"}
    assert kwargs == {"d": {"default": _resources.DEFAULT_ARG_VALUE, "type": "string"}}
    assert body_params == ["a"]
    assert sorted(query_params) == sorted(["c", "d"])
    assert path_params == ["b"]


def test_parse_param():
    param = {
        "name": "A",
        "in": "query",
        "required": True,
        "description": "yeah!",
        "type": "string",
    }
    x = _resources.parse_param(param)
    expected = [
        {
            "in": "query",
            "name": "a",
            "required": True,
            "doc": "a : str\n    yeah!\n",
            "type": "str",
        }
    ]
    assert x == expected


def test_parse_params():
    param = {
        "name": "A",
        "in": "query",
        "required": False,
        "description": "yeah!",
        "type": "string",
    }
    param2 = {
        "name": "B",
        "in": "path",
        "required": True,
        "description": "nah!",
        "type": "integer",
    }
    x, y = _resources.parse_params([param, param2], "summary!", "get", "objects")
    expect_x = [
        {
            "in": "query",
            "doc": "a : str, optional\n    yeah!\n",
            "required": False,
            "name": "a",
            "default": _resources.DEFAULT_ARG_VALUE,
            "type": "str",
        },
        {
            "in": "path",
            "doc": "b : int\n    nah!\n",
            "required": True,
            "name": "b",
            "type": "int",
        },
    ]
    expect_y = (
        "summary!\n\n"
        "API URL: ``GET /objects``\n\n"
        "Parameters\n"
        "----------\n"
        "b : int\n"
        "    nah!\n"
        "a : str, optional\n"
        "    yeah!\n"
    )
    assert x == expect_x
    assert y == expect_y


def test_parse_param_body():
    expected = [
        {
            "required": False,
            "name": "a",
            "in": "body",
            "doc": "a : List, optional\n",
            "default": _resources.DEFAULT_ARG_VALUE,
            "type": "List",
        }
    ]
    param_body = {"schema": {"properties": {"A": {"type": "array"}}}}
    x = _resources.parse_param_body(param_body)
    assert x == expected

    expected_with_default = [
        {
            "required": False,
            "name": "a",
            "in": "body",
            "doc": "a : List, optional\n",
            "default": 50,
            "type": "List",
        }
    ]
    param_body_with_default = {
        "schema": {"properties": {"A": {"type": "array", "default": 50}}}
    }  # noqa: E501
    x_with_default = _resources.parse_param_body(param_body_with_default)
    assert x_with_default == expected_with_default


def test_parse_method_name():
    x = _resources.parse_method_name("get", "url.com/containers")
    y = _resources.parse_method_name("get", "url.com/containers/{id}")
    z = _resources.parse_method_name("get", "url.com/containers/{id}/runs/{run_id}")
    a = _resources.parse_method_name("post", "url.com/containers/{id}/runs/{run_id}")
    b = _resources.parse_method_name("get", "url.com/containers/{id}/{run_id}")
    c = _resources.parse_method_name("get", "url.com/containers/{id}/{run_id}/shares")
    assert x == "list_containers"
    assert y == "get_containers"
    assert z == "get_containers_runs"
    assert a == "post_containers_runs"
    assert b == "get_containers_id"
    assert c == "list_containers_id_shares"


def test_duplicate_names_generated_from_api_spec():
    resolved_civis_api_spec = JsonRef.replace_refs(API_SPEC)
    paths = resolved_civis_api_spec["paths"]
    classes = defaultdict(list)
    for path, ops in paths.items():
        class_name, methods = _resources.parse_path(path, ops, "1.0")
        method_names = [x[0] for x in methods]
        classes[class_name].extend(method_names)
    for cls, names in classes.items():
        err_msg = "Duplicate methods in {}: {}".format(cls, sorted(names))
        assert len(set(names)) == len(names), err_msg


class MockExpiredKeyResponse:
    status_code = 401


mock_str = "civis.resources._resources.requests.Session.get"


@mock.patch(mock_str, return_value=MockExpiredKeyResponse)
def test_expired_api_key(mock_response):
    msg = "401 error downloading API specification. API key may be expired."
    http_error_raised = False
    try:
        _resources.get_api_spec("expired_key", "1.0")
    except HTTPError as err:
        http_error_raised = True
        assert str(err) == msg
    assert http_error_raised


def _create_mock_endpoint():
    args = [
        {"name": "foo", "in": "query", "required": True, "doc": ""},
        {"name": "bar", "in": "query", "required": False, "doc": ""},
    ]
    method = _resources.create_method(
        args,
        "get",
        "mock_name",
        "/objects",
        "deprecation",
        "param_doc",
        "resp_doc",
        "return_annotation",
    )
    mock_endpoint = mock.MagicMock()
    return mock_endpoint, method


def test_create_method_unexpected_kwargs():
    mock_endpoint, method = _create_mock_endpoint()

    # Method works without unexpected kwarg
    method(mock_endpoint, foo=0, bar=0)
    mock_endpoint._call_api.assert_called_once_with(
        "get", "/objects", {"foo": 0, "bar": 0}, {}, "deprecation", iterator=False
    )

    # Method raises TypeError with unexpected kwarg
    expected_msg = "mock_name() got an unexpected keyword argument(s) " "{'baz'}"
    with pytest.raises(TypeError) as excinfo:
        method(mock_endpoint, foo=0, bar=0, baz=0)
    assert str(excinfo.value) == expected_msg


def test_create_method_too_many_pos_args():
    mock_endpoint, method = _create_mock_endpoint()

    # Method raises TypeError with too many arguments
    with pytest.raises(TypeError) as excinfo:
        method(mock_endpoint, 0, 0, 0)
    assert str(excinfo.value) == "too many positional arguments"


def test_create_method_multiple_values():
    mock_endpoint, method = _create_mock_endpoint()

    # Method raises TypeError with multiple values for arguments
    with pytest.raises(TypeError) as excinfo:
        method(mock_endpoint, 0, foo=0)
    assert str(excinfo.value) == "multiple values for argument 'foo'"


def test_create_method_keyword_only():
    # Verify that optional arguments are keyword-only
    # (This language feature is only present in Python 3)
    mock_endpoint, method = _create_mock_endpoint()

    with pytest.raises(TypeError) as excinfo:
        method(mock_endpoint, 0, 0)
    assert str(excinfo.value) == "too many positional arguments"


def test_create_method_deprecation_warning():
    args = [{"name": "foo", "in": "query", "required": True, "doc": ""}]
    method = _resources.create_method(
        args,
        "get",
        "mock_name",
        "/objects",
        "deprecation",
        "param_doc",
        "resp_doc",
        "return_annotation",
    )
    mock_endpoint = Endpoint({"api_key": "abc"}, client=create_client_mock())
    mock_endpoint._make_request = mock.Mock()

    with pytest.warns(FutureWarning, match="deprecation"):
        method(mock_endpoint, foo=0)


@mock.patch("builtins.open", new_callable=mock.mock_open, read_data='{"test": true}')
@mock.patch("civis.resources._resources.generate_classes", autospec=True)
@mock.patch("civis.resources._resources.parse_api_spec", autospec=True)
def test_generate_classes_maybe_cached(mock_parse, mock_gen, mock_open):
    api_key = "mock"
    api_version = "1.0"

    # Calls generate_classes when no cache is passed
    _resources.generate_classes_maybe_cached(None, api_key, api_version)
    mock_gen.assert_called_once_with(api_key, api_version, DEFAULT_USER_AGENT)
    mock_gen.reset_mock()

    # Handles OrderedDict
    spec = OrderedDict({"test": True})
    _resources.generate_classes_maybe_cached(spec, api_key, api_version)
    mock_parse.assert_called_once_with(spec, api_version)
    assert not mock_gen.called

    # Handles str
    mock_parse.reset_mock()
    _resources.generate_classes_maybe_cached("mock", api_key, api_version)
    mock_parse.assert_called_once_with(spec, api_version)
    assert not mock_gen.called

    # Error when a regular dict is passed
    bad_spec = {"test": True}
    with pytest.raises(ValueError):
        _resources.generate_classes_maybe_cached(bad_spec, api_key, api_version)


@mock.patch("civis.resources._resources.parse_method", autospec=True)
def test_parse_api_spec_names(mock_method):
    """Test that path parsing preserves underscore in resource name."""
    mock_method.return_value = ("method_a", lambda x: x)
    mock_ops = {"get": None, "post": None}
    mock_paths = {
        "/two_words/": mock_ops,
        "/oneword/": mock_ops,
        "/hyphen-words": mock_ops,
    }
    mock_api_spec = {"paths": mock_paths}
    classes = _resources.parse_api_spec(mock_api_spec, "test_api_version")
    assert sorted(classes.keys()) == ["hyphen_words", "oneword", "two_words"]
    assert classes["oneword"].__name__ == "Oneword"
    assert classes["two_words"].__name__ == "Two_Words"
    assert classes["hyphen_words"].__name__ == "Hyphen_Words"


def test_client_pyi_matches_api_spec():
    with tempfile.TemporaryDirectory() as temp_dir:
        test_client_pyi_path = os.path.join(temp_dir, "test_client.pyi")
        generate_client_pyi(test_client_pyi_path, API_SPEC_PATH)
        actual = open(CLIENT_PYI_PATH).read()
        expected = open(test_client_pyi_path).read()
        match = expected == actual
        # Avoid the more direct `assert expected == actual`,
        # or else pytest would print the unwieldy, long diffs for a mismatch.
        assert match, (
            "client.pyi doesn't match the API spec in the codebase. "
            "Run tools/update_civis_api_spec.py."
        )


@mock.patch("civis.resources._resources.generate_classes")
def test_generate_classes_with_ttl_cache__base_case(mock_gen):
    # Calling generate_classes_maybe_cached multiple times should only call
    # generate_classes once.
    for _ in range(5):
        _resources.generate_classes_maybe_cached(None, "api-key", "1.0")
    assert mock_gen.call_count == 1


@mock.patch("civis.resources._resources.generate_classes")
def test_generate_classes_with_ttl_cache__force_clear_cache(mock_gen):
    # Check that we can force clear the cache.
    for _ in range(5):
        _resources.generate_classes_maybe_cached(
            None, "api-key", "1.0", force_refresh_api_spec=True
        )
    assert mock_gen.call_count == 5


@mock.patch("civis.resources._resources._spec_expire_time")
@mock.patch("civis.resources._resources.generate_classes")
def test_generate_classes_with_ttl_cache__expire_cache(mock_gen, mock_expire_time):
    # Check that we can expire the cache.
    mock_expire_time.return_value = 0.001
    for _ in range(5):
        _resources.generate_classes_maybe_cached(None, "api-key", "1.0")
        # Sleep the same amount of time as the mock_expire_time,
        # so that we should trigger the cache clearing at every iteration.
        time.sleep(0.001)
    assert mock_gen.call_count == 5


@pytest.mark.parametrize("source, expected", [("ab_cd", "AbCd"), ("", "")])
def test_snake_to_camel(source, expected):
    assert _resources._snake_to_camel(source) == expected



import time
import threading

from civis.base import CivisJobFailure, CivisAsyncResultBase, FAILED, DONE
from civis.response import Response


_MAX_POLLING_INTERVAL = 15


class _ResultPollingThread(threading.Thread):
    """Poll a function until it returns a Response with a DONE state"""

    # Inspired by `threading.Timer`

    def __init__(self, pollable_result):
        super().__init__(daemon=True)
        self.pollable_result = pollable_result
        self.finished = threading.Event()

    def cancel(self):
        """Stop the poller if it hasn't finished yet."""
        self.finished.set()

    def join(self, timeout=None):
        """Shut down the polling when the thread is terminated."""
        self.cancel()
        super().join(timeout=timeout)

    def run(self):
        """Poll until done."""
        while not self.finished.wait(self.pollable_result._next_polling_interval):
            # Spotty internet connectivity can result in polling functions
            # returning None. This treats None responses like responses which
            # have a non-DONE state.
            poller_result = self.pollable_result._check_result()
            if poller_result is not None and poller_result.state in DONE:
                self.finished.set()


class PollableResult(CivisAsyncResultBase):
    """A class for tracking pollable results.

    This class will begin polling immediately upon creation, and poll for
    job completion once every `polling_interval` seconds until the job
    completes in Civis.

    Parameters
    ----------
    poller : func
        A function which returns an object that has a ``state`` attribute.
    poller_args : tuple
        The arguments with which to call the poller function.
    polling_interval : int or float, optional
        The number of seconds between API requests to check whether a result
        is ready. If an integer or float is provided, this number will be used
        as the polling interval. If ``None`` (the default), the polling interval will
        start at 1 second and increase geometrically up to 15 seconds. The ratio of
        the increase is 1.2, resulting in polling intervals in seconds of
        1, 1.2, 1.44, 1.728, etc. This default behavior allows for a faster return for
        a short-running job and a capped polling interval for longer-running jobs.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    poll_on_creation : bool, optional
        If ``True`` (the default), it will poll upon calling ``result()`` the
        first time. If ``False``, it will wait the number of seconds specified
        in `polling_interval` from object creation before polling.

    Examples
    --------
    >>> client = civis.APIClient()
    >>> database_id = client.get_database_id("my_database")
    >>> cred_id = client.default_database_credential_id
    >>> sql = "SELECT 1"
    >>> preview_rows = 10
    >>> response = client.queries.post(database_id, sql, preview_rows,
    >>>                                credential=cred_id)
    >>> job_id = response.id
    >>>
    >>> poller = client.queries.get
    >>> poller_args = (job_id, ) # (job_id, run_id) if poller requires run_id
    >>> polling_interval = 10
    >>> poll = PollableResult(poller, poller_args, polling_interval)
    """

    # this may not be friendly to a rate-limited api
    # Implementation notes: The `PollableResult` depends on some private
    # features of the `concurrent.futures.Future` class, so it's possible
    # that future versions of Python could break something here.
    # (It works under at least 3.6)
    # We use the following `Future` implementation details
    # - The `Future` checks its state against predefined strings. We use
    #   `STATE_TRANS` to translate from the Civis platform states to `Future`
    #    states.
    # - `Future` uses a `_state` attribute to check its current condition
    # - `Future` handles event notification through `set_result` and
    #   `set_exception`, which we call from `_check_result`.
    # - We use the `Future` thread lock called `_condition`
    # - We assume that results of the Future are stored in `_result`.
    def __init__(
        self,
        poller,
        poller_args,
        polling_interval=None,
        client=None,
        poll_on_creation=True,
    ):
        super().__init__()

        self.poller = poller
        self.poller_args = poller_args
        self.polling_interval = polling_interval
        self.client = client
        self.poll_on_creation = poll_on_creation

        if self.polling_interval is not None and self.polling_interval <= 0:
            raise ValueError("The polling interval must be positive.")

        self._next_polling_interval = 1
        self._use_geometric_polling = True

        # Polling arguments. Never poll more often than the requested interval.
        if poll_on_creation:
            self._last_polled = None
        else:
            self._last_polled = time.time()
        self._last_result = None

        self._begin_tracking()

    def _begin_tracking(self, start_thread=False):
        """Start monitoring the Civis Platform job"""
        with self._condition:
            if getattr(self, "poller", None) is None:
                raise RuntimeError(
                    "Internal error: Must set polling "
                    "function before initializing thread."
                )
            self._reset_polling_thread(self.polling_interval, start_thread)

    def _check_result(self):
        """Return the job result from Civis. Once the job completes, store the
        result and never poll again."""
        with self._condition:
            # Start a single thread continuously polling.
            # It will stop once the job completes.
            if not self._polling_thread.is_alive() and self._result is None:
                self._polling_thread.start()

            if self._result is not None:
                # If the job is already completed, just return the stored
                # result.
                return self._result

            # Check to see if the job has finished, but don't poll more
            # frequently than the requested polling frequency.
            now = time.time()
            if (
                not self._last_polled
                or (now - self._last_polled) >= self._next_polling_interval
            ):
                if self._use_geometric_polling:
                    # Choosing a common ratio of 1.2 for these polling intervals:
                    #   1, 1.2, 1.44, 1.73, 2.07, 2.49, 2.99, ..., and capped at 15.
                    # Within the first 15 secs by wall time, we call the poller 7 times,
                    # which gives a short-running job's future.result()
                    # a higher chance to return faster.
                    # For longer running jobs, the polling interval will be capped
                    # at 15 secs when by wall time 87 secs have passed.
                    self._next_polling_interval *= 1.2
                    if self._next_polling_interval > _MAX_POLLING_INTERVAL:
                        self._next_polling_interval = _MAX_POLLING_INTERVAL
                        self._use_geometric_polling = False
                # Poll for a new result
                self._last_polled = now
                try:
                    self._last_result = self.poller(*self.poller_args)
                except Exception as e:
                    # The _poller can raise API exceptions
                    # Set those directly as this Future's exception
                    self._set_api_exception(exc=e)
                else:
                    # If the job has finished, then register completion and
                    # store the results. Because of the `if self._result` check
                    # up top, we will never get here twice.
                    self._set_api_result(self._last_result)

            return self._last_result

    def _set_api_result(self, result):
        with self._condition:
            if result.state in FAILED:
                try:
                    err_msg = str(result["error"])
                except:  # NOQA
                    err_msg = str(result)
                job_id = getattr(self, "job_id", None)
                run_id = getattr(self, "run_id", None)
                self._set_api_exception(
                    exc=CivisJobFailure(err_msg, result, job_id, run_id),
                    result=result,
                )
            elif result.state in DONE:
                self.set_result(result)
                self.cleanup()

    def _set_api_exception(self, exc, result=None):
        with self._condition:
            if result is None:
                result = Response({"state": FAILED[0]})
            self._result = result
            self._last_result = self._result
            self.set_exception(exc)
            self.cleanup()

    def cleanup(self):
        # This gets called after the result is set.
        # Ensure that the polling thread shuts down when it's no longer needed.
        with self._condition:
            if self._polling_thread.is_alive():
                self._polling_thread.cancel()

    def _reset_polling_thread(self, polling_interval, start_thread=False):
        with self._condition:
            if (
                getattr(self, "_polling_thread", None) is not None
                and self._polling_thread.is_alive()
            ):
                self._polling_thread.cancel()
            self.polling_interval = polling_interval
            self._next_polling_interval = 1 if (pi := polling_interval) is None else pi
            self._use_geometric_polling = polling_interval is None
            self._polling_thread = _ResultPollingThread(self)
            if start_thread:
                self._polling_thread.start()



"""Mock client creation and tooling"""

import warnings
from functools import lru_cache
from unittest import mock

from civis import APIClient
from civis.resources import API_SPEC_PATH
from civis.response import Response


def create_client_mock(cache=None):
    """Create an APIClient mock from a cache of the API spec

    Parameters
    ----------
    cache : str, optional
        Location of the API spec on the local filesystem.
        If ``None`` or not given, the default API spec will be used.

    Returns
    -------
    mock.Mock
        A `Mock` object which looks like an APIClient and which will
        error if any method calls have non-existent / misspelled parameters
    """
    if cache is None:
        cache = API_SPEC_PATH

    # Create a client from the cache. We'll use this for auto-speccing.
    real_client = _real_client(cache)

    # Prevent the client from trying to talk to the real API when autospeccing
    with mock.patch("requests.Session", mock.MagicMock), warnings.catch_warnings():
        # Ignore deprecation warning from `client.default_credential`.
        warnings.simplefilter("ignore", FutureWarning)
        mock_client = mock.create_autospec(real_client, spec_set=True)

    return mock_client


def create_client_mock_for_container_tests(
    script_id=-10, run_id=100, state="succeeded", run_outputs=None, log_outputs=None
):
    """Returns a CivisAPIClient Mock set up for testing methods that use
    container scripts. Contains endpoint method mocks and return values
    for posting container jobs, retrieving outputs, and reading logs.
    Also contains the mocks to cancel the container when the state
    is set to 'failed'.

    Parameters
    ----------
    script_id: int
        Mock-create containers with this ID when calling `post_containers`
        or `post_containers_runs`.
    run_id: int
        Mock-create runs with this ID when calling `post_containers_runs`.
    state: str, optional
        The reported state of the container run
    run_outputs: list, optional
        List of Response objects returned as run outputs
    log_outputs : list, optional
        List of Response objects returned as log outputs

    Returns
    -------
    `unittest.mock.Mock`
        With scripts endpoints `post_containers`, `post_containers_runs`,
        `post_cancel`, and `get_containers_runs` set up.
    """
    c = create_client_mock()

    mock_container = Response({"id": script_id})
    c.scripts.post_containers.return_value = mock_container
    mock_container_run_start = Response(
        {"id": run_id, "container_id": script_id, "state": "queued"}
    )
    mock_container_run_json = {"id": run_id, "container_id": script_id, "state": state}
    if state == "failed":
        mock_container_run_json["error"] = "None"
    mock_container_run = Response(mock_container_run_json)
    c.scripts.post_containers_runs.return_value = mock_container_run_start
    c.scripts.get_containers_runs.return_value = mock_container_run
    c.scripts.list_containers_runs_outputs.return_value = run_outputs or []
    c.jobs.list_runs_outputs.return_value = run_outputs or []
    c.jobs.list_runs_logs.return_value = log_outputs or []

    def change_state_to_cancelled(script_id):
        mock_container_run_json = mock_container_run.json()
        mock_container_run_json["state"] = "cancelled"
        return Response(mock_container_run_json)

    c.scripts.post_cancel.side_effect = change_state_to_cancelled

    return c


@lru_cache(maxsize=1)
def _real_client(local_api_spec):
    real_client = APIClient(local_api_spec=local_api_spec, api_key="none")
    real_client._feature_flags = {"noflag": None}
    return real_client



import logging
import operator
import time
from datetime import datetime

from civis import APIClient
from civis.futures import CivisFuture

log = logging.getLogger(__name__)

_FOLLOW_POLL_INTERVAL_SEC = 5
_LOG_REFETCH_CUTOFF_SECONDS = 300
_LOG_REFETCH_COUNT = 100
_LOGS_PER_QUERY = 250


def run_job(job_id, client=None, polling_interval=None):
    """Run a job.

    Parameters
    ----------
    job_id: str or int
        The ID of the job.
    client: :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    polling_interval : int or float, optional
        The number of seconds between API requests to check whether a result
        is ready.

    Returns
    -------
    results: :class:`~civis.futures.CivisFuture`
        A `CivisFuture` object.
    """
    if client is None:
        client = APIClient()
    run = client.jobs.post_runs(job_id)
    return CivisFuture(
        client.jobs.get_runs,
        (job_id, run["id"]),
        client=client,
        polling_interval=polling_interval,
        poll_on_creation=False,
    )


def run_template(id, arguments, JSONValue=False, client=None):
    """Run a template and return the results.

    Parameters
    ----------
    id: int
        The template id to be run.
    arguments: dict
        Dictionary of arguments to be passed to the template.
    JSONValue: bool, optional
        If True, will return the JSON output of the template.
        If False, will return the file ids associated with the
        output results.
    client: :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.

    Returns
    -------
    output: dict
        If JSONValue = False, dictionary of file ids with the keys
        being their output names.
        If JSONValue = True, JSON dict containing the results of the
        template run. Expects only a single JSON result. Will return
        nothing if either there is no JSON result or there is more
        than 1 JSON result.

    Examples
    --------
    >>> # Run template to return file_ids
    >>> run_template(my_template_id, arguments=my_dict_of_args)
    {'output': 1234567}
    >>> # Run template to return JSON output
    >>> run_template(my_template_id, arguments=my_dict_of_args, JSONValue=True)
    {'result1': 'aaa', 'result2': 123}
    """
    if client is None:
        client = APIClient()
    job = client.scripts.post_custom(id, arguments=arguments)
    run = client.scripts.post_custom_runs(job.id)
    fut = CivisFuture(client.scripts.get_custom_runs, (job.id, run.id), client=client)
    fut.result()
    outputs = client.scripts.list_custom_runs_outputs(job.id, run.id)
    if JSONValue:
        json_output = [o.value for o in outputs if o.object_type == "JSONValue"]
        if len(json_output) == 0:
            log.warning("No JSON output for template {}".format(id))
            return
        if len(json_output) > 1:
            log.warning(
                "More than 1 JSON output for template {}"
                " -- returning only the first one.".format(id)
            )
        # Note that the cast to a dict is to convert
        # an expected Response object.
        return json_output[0].json()
    else:
        file_ids = {o.name: o.object_id for o in outputs}
        return file_ids


def _timestamp_from_iso_str(s):
    """Return an integer POSIX timestamp for a given ISO date string.

    Note: Until Python 3.11, datetime.fromisoformat doesn't work
    with the format returned by Civis Platform.
    """
    try:
        return datetime.fromisoformat(s).timestamp()
    except ValueError:
        try:
            # This is the format that Civis Platform returns.
            return datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%f%z").timestamp()
        except ValueError:
            # Another format, just in case.
            return datetime.strptime(s, "%Y-%m-%dT%H:%M:%S%z").timestamp()


def _compute_effective_max_log_id(logs):
    """Find a max log ID use in order to avoid missing late messages.

    The order of log IDs may not be consistent with "created at" times
    since log entries are created by Civis Platform as well as the code
    for the job itself. This function looks through recent logs
    and finds a maximum ID that is at least as old as a set cutoff period,
    so that messages with lower IDs that show up a bit late won't be skipped.
    With this, it is still theoretically possible but extremely unlikely
    for some late log messages to be skipped in the job_logs function.
    """
    if not logs:
        return 0

    sorted_logs = sorted(logs, key=operator.itemgetter("id"))

    max_created_at_timestamp = _timestamp_from_iso_str(sorted_logs[-1]["created_at"])
    cutoff = time.time() - _LOG_REFETCH_CUTOFF_SECONDS
    if max_created_at_timestamp < cutoff:
        return sorted_logs[-1]["id"]
    elif len(sorted_logs) >= _LOG_REFETCH_COUNT:
        return sorted_logs[-_LOG_REFETCH_COUNT]["id"]

    return 0


def _job_finished_past_timeout(job_id, run_id, finished_timeout, client):
    """Return true if the run finished more than so many seconds ago."""
    if finished_timeout is None:
        return False

    run = client.jobs.get_runs(job_id, run_id)
    finished_at = run.json()["finished_at"]
    if finished_at is None:
        return False
    finished_at_ts = _timestamp_from_iso_str(finished_at)
    result = finished_at_ts < time.time() - finished_timeout
    return result


def job_logs(job_id, run_id=None, finished_timeout=None, client=None):
    """Return a generator of log message dictionaries for a given run.

    Parameters
    ----------
    job_id : int
        The ID of the job to retrieve log message for.
    run_id : int or None
        The ID of the run to retrieve log messages for.
        If None, the ID for the most recent run will be used.
    finished_timeout: int or None
        If not None, then this function will return once the run has
        been finished for the specified number of seconds.
        If None, then this function will wait until the API says there
        will be no more new log messages, which may take a few minutes.
        A timeout of 30-60 seconds is usually enough to retrieve all
        log messages.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.

    Yields
    ------
    dict
        A log message dictionary with "message", "created_at" and other attributes
        provided by the job logs endpoint. Note that this will block execution
        until the job has stopped and all log messages are retrieved.

    Examples
    --------
    >>> # Print all log messages from a job's most recent run
    >>> for log in job_logs(job_id=123456):
    ...     print(f"{log['created_at']}: {log['message']}")
    ...
    >>> # Get logs from a specific run with a 30 second timeout
    >>> for log in job_logs(job_id=123456, run_id=789, finished_timeout=30):
    ...     print(log['message'])
    """
    client = client or APIClient()

    if run_id is None:
        run_id = client.jobs.list_runs(
            job_id, limit=1, order="id", order_dir="desc"
        ).json()[0]["id"]

    local_max_log_id = 0
    continue_polling = True

    known_log_ids = set()

    while continue_polling:
        # This call gets a limited number of log messages since last_id,
        # ordered by log ID.
        response = client.jobs.list_runs_logs(
            job_id,
            run_id,
            last_id=local_max_log_id,
            limit=_LOGS_PER_QUERY,
        )
        if "civis-max-id" in response.headers:
            remote_max_log_id = int(response.headers["civis-max-id"])
        else:
            # Platform hasn't seen any logs at all yet
            remote_max_log_id = None
        logs = response.json()
        if logs:
            local_max_log_id = max(log["id"] for log in logs)
            logs.sort(key=operator.itemgetter("created_at", "id"))
        for log in logs:
            if log["id"] in known_log_ids:
                continue
            known_log_ids.add(log["id"])
            yield log

        log_finished = response.headers["civis-cache-control"] != "no-store"

        if remote_max_log_id is None:
            remote_has_more_logs_to_get_now = False
        elif local_max_log_id == remote_max_log_id:
            remote_has_more_logs_to_get_now = False
            local_max_log_id = _compute_effective_max_log_id(logs)
            if log_finished or _job_finished_past_timeout(
                job_id, run_id, finished_timeout, client
            ):
                continue_polling = False
        else:
            remote_has_more_logs_to_get_now = True

        if continue_polling and not remote_has_more_logs_to_get_now:
            time.sleep(_FOLLOW_POLL_INTERVAL_SEC)



import os
import re

import civis


_REPO_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))


def test_version_number_match_with_changelog():
    """__version__ and CHANGELOG.md match for the latest version number."""
    changelog = open(os.path.join(_REPO_DIR, "CHANGELOG.md")).read()
    version_in_changelog = re.search(r"##\s+(\d+\.\d+\.\d+)", changelog).groups()[0]
    assert civis.__version__ == version_in_changelog, (
        "Make sure both __version__ and CHANGELOG are updated to match the "
        "latest version number"
    )
